/*
 * Code borrowed from powerpc/kernel/pci-common.c
 *
 * Copyright (C) 2003 Anton Blanchard <anton@au.ibm.com>, IBM
 * Copyright (C) 2014 ARM Ltd.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 */

#include <linux/init.h>
#include <linux/io.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/of_pci.h>
#include <linux/of_platform.h>
#include <linux/slab.h>
#include <linux/acpi.h>
#include <linux/pci-acpi.h>

#include <asm/pci-bridge.h>

/*
 * Called after each bus is probed, but before its children are examined
 */
void pcibios_fixup_bus(struct pci_bus *bus)
{
	/* nothing to do, expected to be removed in the future */
}

/*
 * We don't have to worry about legacy ISA devices, so nothing to do here
 */
resource_size_t pcibios_align_resource(void *data, const struct resource *res,
				resource_size_t size, resource_size_t align)
{
	return res->start;
}

int pcibios_root_bridge_prepare(struct pci_host_bridge *bridge)
{
	struct pci_sysdata *sd;

	if (!acpi_disabled) {
		sd = bridge->bus->sysdata;
		ACPI_COMPANION_SET(&bridge->dev, sd->companion);
	}
	return 0;
}

/*
 * Try to assign the IRQ number from DT when adding a new device
 */
int pcibios_add_device(struct pci_dev *dev)
{
	if (acpi_disabled)
		dev->irq = of_irq_parse_and_map_pci(dev, 0, 0);

	return 0;
}

void pcibios_add_bus(struct pci_bus *bus)
{
	if (!acpi_disabled)
		acpi_pci_add_bus(bus);
}

void pcibios_remove_bus(struct pci_bus *bus)
{
	if (!acpi_disabled)
		acpi_pci_remove_bus(bus);
}

int pcibios_enable_irq(struct pci_dev *dev)
{
	if (!acpi_disabled && !pci_dev_msi_enabled(dev))
		acpi_pci_irq_enable(dev);
	return 0;
}

int pcibios_disable_irq(struct pci_dev *dev)
{
	if (!acpi_disabled && !pci_dev_msi_enabled(dev))
		acpi_pci_irq_disable(dev);
	return 0;
}

int pcibios_enable_device(struct pci_dev *dev, int bars)
{
	int err;

	err = pci_enable_resources(dev, bars);
	if (err < 0)
		return err;

	if (!pci_dev_msi_enabled(dev))
		return pcibios_enable_irq(dev);
	return 0;
}

#ifdef CONFIG_PCI_DOMAINS_GENERIC
void pci_bus_assign_domain_nr(struct pci_bus *bus, struct device *parent)
{
	int domain = -1;

	if (acpi_disabled)
		domain = of_get_pci_domain_nr(parent->of_node);
	else {
		struct pci_sysdata *sd = bus->sysdata;

		domain = sd->domain;
	}
	if (domain >= 0)
		bus->domain_nr = domain;
}
#endif

static int __init pcibios_assign_resources(void)
{
	struct pci_bus *root_bus;

	if (acpi_disabled)
		return 0;

	list_for_each_entry(root_bus, &pci_root_buses, node) {
		pcibios_resource_survey_bus(root_bus);
		pci_assign_unassigned_root_bus_resources(root_bus);
	}
	return 0;
}
/*
 * fs_initcall comes after subsys_initcall, so we know acpi scan
 * has run.
 */
fs_initcall(pcibios_assign_resources);
